from typing import ClassVar

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract import AbstractDetector

from mxcubeweb.core.adapter.adapter_base import AdapterBase
from mxcubeweb.core.models.configmodels import ResourceHandlerConfigModel

resource_handler_config = ResourceHandlerConfigModel(
    commands=["display_image"],
    attributes=["data", "get_value"],
)


class DetectorAdapter(AdapterBase):
    SUPPORTED_TYPES: ClassVar[list[object]] = [AbstractDetector.AbstractDetector]

    def __init__(self, ho, role, app):
        """
        Args:
            (object): Hardware object.
        """
        super().__init__(ho, role, app, resource_handler_config)
        ho.connect("stateChanged", self._state_change)

    def get_value(self) -> dict:
        """
        Get the file suffix of the data files.
        """
        return {"fileSuffix": HWR.beamline.detector.get_property("file_suffix", "?")}

    def _state_change(self, *args, **kwargs):
        self.state_change(*args, **kwargs)

    def state(self):
        return self._ho.get_state().name.upper()

    def display_image(self, path: str, img_num) -> dict:
        """
        Notify ADXV and/or Braggy of the image to display.
        """
        res = {"image_url": ""}

        if path:
            fpath, img = HWR.beamline.detector.get_actual_file_path(path, img_num)
            HWR.beamline.collect.adxv_notify(fpath, img)
            fpath = HWR.beamline.session.get_path_with_proposal_as_root(fpath)

            if self.app.config.braggy.USE_BRAGGY:
                res = {
                    "image_url": (
                        f"{self.app.config.braggy.BRAGGY_URL}/"
                        f"?file={fpath}/image_${img_num}.h5.dataset"
                    )
                }

        return res



import json
import logging

import gevent
from mxcubecore import HardwareRepository as HWR
from mxcubecore import queue_entry
from mxcubecore.HardwareObjects.abstract.AbstractSampleChanger import SampleChangerState

from mxcubeweb.core.components.component_base import ComponentBase


# TO CONSIDER:
# This should maybe be made into an adapter instead of a component
class SampleChanger(ComponentBase):
    def __init__(self, app, config):
        super().__init__(app, config)

    def init_signals(self):
        sc = HWR.beamline.sample_changer

        # Initialize hwobj signals
        sc.connect("stateChanged", self._sc_state_changed)
        sc.connect("isCollisionSafe", self._is_collision_safe)
        sc.connect("loadedSampleChanged", self._loaded_sample_changed)
        sc.connect("contentsUpdated", self._sc_contents_update)

        if HWR.beamline.sample_changer_maintenance is not None:
            HWR.beamline.sample_changer_maintenance.connect(
                "globalStateChanged", self._sc_maintenance_update
            )

            HWR.beamline.sample_changer_maintenance.connect(
                "gripperChanged", self._gripper_changed
            )

    def _sc_state_changed(self, *args):
        new_state = args[0]
        state_str = SampleChangerState.STATE_DESC.get(new_state, "Unknown").upper()
        self.app.server.emit("sc_state", state_str, namespace="/hwr")

    def _is_collision_safe(self, *args):
        new_state = args[0]
        # we are only interested when it becames true
        if new_state:
            msg = {
                "signal": "isCollisionSafe",
                "message": "Sample moved to safe area",
            }
            self.app.server.emit("sc", msg, namespace="/hwr")

    def _loaded_sample_changed(self, sample):
        if hasattr(sample, "get_address"):
            address = sample.get_address()
            barcode = sample.get_id()
        else:
            address = ""
            barcode = ""

        logging.getLogger("HWR").info("Loaded sample changed: %s", address)

        try:
            sampleID = address

            if HWR.beamline.sample_changer.has_loaded_sample():
                self.app.lims.set_current_sample(sampleID)
            else:
                sample = HWR.beamline.sample_changer.get_loaded_sample()
                address = sample.get_address() if sample else None
                self.app.lims.set_current_sample(address)

            self.app.server.emit(
                "loaded_sample_changed",
                {"address": address, "barcode": barcode},
                namespace="/hwr",
            )

            self._sc_load_ready(address)
        except Exception:
            logging.getLogger("HWR").exception("Error setting loaded sample")

    def _sc_load_ready(self, location):
        msg = {
            "signal": "loadReady",
            "location": location,
            "message": "Sample changer, loaded sample",
        }

        self.app.server.emit("sc", msg, namespace="/hwr")

    def _sc_unload(self, location):
        msg = {
            "signal": "operatingSampleChanger",
            "location": location,
            "message": "Please wait, unloading sample",
        }

        self.app.server.emit("sc", msg, namespace="/hwr")

    def _sc_contents_update(self):
        self.app.server.emit("sc_contents_update", {}, namespace="/hwr")

    def _sc_maintenance_update(self, *args):
        if len(args) == 3:
            # Restore the `global_state` parameter removed in this commit 337efd37
            global_state, cmd_state, message = args
        else:
            # Be backward compatible with HW objects which are emitting signal with
            # 2 arguments
            global_state = {}
            cmd_state, message = args

        try:
            self.app.server.emit(
                "sc_maintenance_update",
                {
                    "global_state": global_state,
                    "commands_state": json.dumps(cmd_state),
                    "message": message,
                },
                namespace="/hwr",
            )
        except Exception:
            logging.getLogger("HWR").exception("error sending message")

    def _gripper_changed(self):
        self.app.queue.queue_clear()
        self.app.server.emit(
            "queue", {"Signal": "update", "message": "all"}, namespace="/hwr"
        )

    def get_sc_contents(self):  # noqa: C901
        def _getElementStatus(e):
            if e.is_leaf():
                if e.is_loaded():
                    return "Loaded"
                if e.has_been_loaded():
                    return "Used"
            if e.is_present():
                return "Present"
            return ""

        def _getElementID(e):
            if e == HWR.beamline.sample_changer:
                if e.get_token() is not None:
                    return e.get_token()
            else:
                if e.get_id() is not None:
                    return e.get_id()
            return ""

        def _addElement(parent, element):
            new_element = {
                "name": element.get_address(),
                "status": _getElementStatus(element),
                "id": _getElementID(element),
                "selected": element.is_selected(),
            }

            parent.setdefault("children", []).append(new_element)

            if not element.is_leaf():
                for e in element.get_components():
                    _addElement(new_element, e)

        if HWR.beamline.sample_changer:
            root_name = HWR.beamline.sample_changer.get_address()

            contents = {"name": root_name}

            if hasattr(HWR.beamline.sample_changer, "get_room_temperature_mode"):
                contents["room_temperature_mode"] = (
                    HWR.beamline.sample_changer.get_room_temperature_mode()
                )

            for element in HWR.beamline.sample_changer.get_components():
                if element.is_present():
                    _addElement(contents, element)
        else:
            contents = {"name": "OFFLINE"}

        return contents

    def _mount_sample(self, sample):
        sc = HWR.beamline.sample_changer
        res = False

        try:
            msg = {
                "signal": "operatingSampleChanger",
                "location": sample["location"],
                "message": "Please wait, loading sample",
            }

            self.app.server.emit("sc", msg, namespace="/hwr")

            sid = self.app.lims.get_current_sample().get("sampleID", False)
            current_queue = self.app.queue.queue_to_dict()

            if sample["location"] != "Manual":
                msg = "Mounting sample: %s (%s)" % (
                    sample["location"],
                    sample.get("sampleName", ""),
                )
                logging.getLogger("user_level_log").info(msg)

                if (
                    not sc.get_loaded_sample()
                    or sc.get_loaded_sample().get_address() != sample["location"]
                ):
                    res = sc.load(sample["sampleID"], wait=True)

                if (
                    res
                    and HWR.beamline.queue_manager.centring_method
                    == queue_entry.CENTRING_METHOD.LOOP
                    and not HWR.beamline.diffractometer.in_plate_mode()
                    and not self.app.harvester.mount_from_harvester()
                ):
                    HWR.beamline.diffractometer.reject_centring()
                    msg = "Starting autoloop centring ..."
                    logging.getLogger("MX3.HWR").info(msg)
                    C3D_MODE = HWR.beamline.diffractometer.C3D_MODE
                    HWR.beamline.diffractometer.start_centring_method(C3D_MODE)
                elif HWR.beamline.diffractometer.in_plate_mode():
                    msg = "Starting autoloop Focusing ..."
                    logging.getLogger("MX3.HWR").info(msg)
                    sc.move_to_crystal_position(None)

            else:
                msg = "Mounting sample: %s" % sample["sampleName"]
                logging.getLogger("user_level_log").info(msg)

                self.app.lims.set_current_sample(sample["sampleID"])
                res = True

        except Exception as ex:
            logging.getLogger("MX3.HWR").exception("[SC] sample could not be mounted")

            raise RuntimeError(str(ex)) from ex
        else:
            # Clean up if the new sample was mounted or the current sample was
            # unmounted and the new one, for some reason, failed to mount
            if res or (not res and not sc.get_loaded_sample()):
                HWR.beamline.sample_view.clear_all()

                # We remove the current sample from the queue, if we are moving
                # from one sample to another and the current sample is in the queue

                if sid and current_queue.get(sid, False):
                    node_id = current_queue[sid]["queueID"]
                    self.app.queue.set_enabled_entry(node_id, False)
                    self.app.queue.queue_toggle_sample(
                        self.app.queue.get_entry(node_id)[1]
                    )
        finally:
            self._sc_load_ready(sample["location"])

        return res

    def _unmount_sample(self, sample):
        try:
            self._sc_unload(sample["location"])

            if sample["location"] != "Manual":
                HWR.beamline.sample_changer.unload(sample["location"], wait=False)
            else:
                self.app.lims.set_current_sample(None)
                self._sc_load_ready(sample["location"])

            msg = "[SC] unmounted %s" % sample["location"]
            logging.getLogger("MX3.HWR").info(msg)
        except Exception:
            msg = "[SC] sample could not be mounted"
            logging.getLogger("MX3.HWR").exception(msg)
            raise
        else:
            HWR.beamline.queue_model.mounted_sample = ""
            HWR.beamline.sample_view.clear_all()

    def mount_sample(self, sample, wait=True):  # noqa: FBT002
        if wait:
            self._mount_sample(sample)
        else:
            gevent.spawn(self._mount_sample, sample)

        return self.get_sc_contents()

    def unmount_current(self):
        sc_sample = HWR.beamline.sample_changer.get_loaded_sample()
        if sc_sample:
            location = sc_sample.get_address()
            self._unmount_sample({"location": location})
        else:
            self._unmount_sample({"location": "Manual"})

        return self.get_sc_contents()

    def get_loaded_sample(self):
        try:
            sample = HWR.beamline.sample_changer.get_loaded_sample()
        except Exception:
            logging.getLogger("MX3.HWR").exception("")
            sample = None

        if sample is not None:
            address = sample.get_address()
            barcode = sample.get_id()
        else:
            address = ""
            barcode = ""

        return address, barcode

    def get_capacity(self):
        baskets = HWR.beamline.sample_changer.get_basket_list()
        num_samples = 0
        for basket in baskets:
            num_samples += basket.get_number_of_samples()
        return {
            "num_baskets": len(baskets),
            "num_samples": num_samples,
        }

    def get_maintenance_cmds(self):
        if HWR.beamline.sample_changer_maintenance is not None:
            ret = HWR.beamline.sample_changer_maintenance.get_cmd_info()
        else:
            ret = "SC maintenance controller not defined"

        return ret

    def get_global_state(self):
        try:
            return HWR.beamline.sample_changer_maintenance.get_global_state()
        except Exception:
            logging.getLogger("MX3.HWR").exception("Could not get sc global state")
            return (
                {},
                "SC maintenance controller doesn't response",
                "Can't retrieve the global state",
            )

    def get_initial_state(self):
        if HWR.beamline.sample_changer_maintenance is not None:
            global_state, cmdstate, msg = self.get_global_state()

            cmds = HWR.beamline.sample_changer_maintenance.get_cmd_info()

        else:
            global_state = {}
            cmdstate = "SC maintenance controller not defined"
            cmds = []
            msg = ""

        contents = self.get_sc_contents()
        address, barcode = self.get_loaded_sample()

        loaded_sample = {"address": address, "barcode": barcode}

        try:
            state = HWR.beamline.sample_changer.get_status().upper()
        except Exception:
            state = "OFFLINE"

        return {
            "state": state,
            "loaded_sample": loaded_sample,
            "contents": contents,
            "global_state": {
                "global_state": global_state,
                "commands_state": cmdstate,
            },
            "cmds": {"cmds": cmds},
            "msg": msg,
            "plate_mode": HWR.beamline.diffractometer.in_plate_mode(),
        }

    def sync_with_crims(self):
        """
        To be use mostly when Diffractometer is in plate mode
        This retun a List of crystal dict available in Crims that have been Harvested
        With this user can visualize easier where the crystal are in Plate GUI
        """
        xtal_list = []
        try:
            processing_plan = HWR.beamline.sample_changer.sync_with_crims()
            for x in processing_plan.plate.xtal_list:
                response = {
                    "crystal_uuid": x.crystal_uuid,
                    "row": x.row,
                    "column": x.column,
                    "shelf": x.shelf,
                    "offset_x": x.offset_x,
                    "offset_y": x.offset_y,
                    "image_url": x.image_url,
                    "image_date": x.image_date,
                    "sample": x.sample,
                }
                xtal_list.append(response)
            return {"xtal_list": xtal_list}
        except Exception:
            logging.getLogger("MX3.HWR").exception("Could not get crystal List")
            return {"xtal_list": xtal_list}



"""
Module that contains application wide settings and state as well as functions
for accessing and manipulating those.
"""

import logging
import os
import sys
import time
from logging import StreamHandler
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path

try:
    import graypy
except ImportError:
    graypy = None

from mxcubecore import (
    ColorFormatter,
)
from mxcubecore import HardwareRepository as HWR
from mxcubecore import (
    removeLoggingHandlers,
)

from mxcubeweb.core.adapter.adapter_manager import HardwareObjectAdapterManager
from mxcubeweb.core.components.chat import Chat
from mxcubeweb.core.components.component_base import import_component
from mxcubeweb.core.components.harvester import Harvester
from mxcubeweb.core.components.lims import Lims
from mxcubeweb.core.components.queue import Queue
from mxcubeweb.core.components.samplechanger import SampleChanger
from mxcubeweb.core.components.sampleview import SampleView
from mxcubeweb.core.components.workflow import Workflow
from mxcubeweb.core.components.log import Log
from mxcubeweb.core.models.configmodels import UIComponentModel
from mxcubeweb.logging_handler import MX3LoggingHandler
from mxcubeweb.core.server.resource_handler import ResourceHandlerFactory

removeLoggingHandlers()


class MXCUBEApplication:
    t0 = time.time()
    # Below variables used for internal application state

    # SampleID and sample data of currently mounted sample, to handle samples
    # that are not mounted by sample changer.
    CURRENTLY_MOUNTED_SAMPLE = ""

    # Sample location of sample that are in process of being mounted
    SAMPLE_TO_BE_MOUNTED = ""

    # Look up table for finding the limsID for a corresponding queueID (QueueNode)
    NODE_ID_TO_LIMS_ID = {}

    # Initial file list for user, initialized at login, for creating automatic
    # run numbers
    INITIAL_FILE_LIST = []

    # Lookup table for sample changer location to data matrix or
    # data matrix to location
    SC_CONTENTS = {"FROM_CODE": {}, "FROM_LOCATION": {}}

    # Current sample list, with tasks
    SAMPLE_LIST = {"sampleList": {}, "sampleOrder": []}

    # Users currently logged in
    USERS = {}

    # Path to video device (i.e. /dev/videoX)
    VIDEO_FORMAT = "MPEG1"

    # Contains the complete client side ui state, managed up state_storage.py
    UI_STATE = {}
    TEMP_DISABLED = []

    # Below variables used for application wide settings

    # Enabled or Disable remote usage
    ALLOW_REMOTE = False

    # Enable timeout gives control (if ALLOW_REMOTE is True)
    TIMEOUT_GIVES_CONTROL = False

    # Enable automatic Mountie of sample when queue executed in
    # "automatic/pipeline" mode
    AUTO_MOUNT_SAMPLE = False

    # Automatically add and execute diffraction plans coming from
    # characterizations
    AUTO_ADD_DIFFPLAN = False

    # Number of sample snapshots taken before collect
    DEFAULT_NUM_SNAPSHOTS = 4

    # Remember collection paramters between samples
    # or reset (defualt) between samples.
    REMEMBER_PARAMETERS_BETWEEN_SAMPLES = False

    CONFIG = None

    mxcubecore = None

    server = None

    def __init__(self):
        msg = "MXCUBEApplication is to be used as a pure static class, dont instanciate"
        raise NotImplementedError(msg)

    @staticmethod
    def init(
        server,
        allow_remote,
        ra_timeout,
        log_fpath,
        log_level,
        enabled_logger_list,
        cfg,
    ):
        """
        Initializes application wide variables, sample video stream, and applies

        :param hwr: HardwareRepository module
        :param bool allow_remote: Allow remote usage, True else False
        :param bool ra_timeout: Timeout gives control, True else False

        :return None:
        """
        # The routes created by the AdapterResourceHandler
        # via the factory are kept between calls to init as they
        # are stored in class variable and only initialized once
        #
        # This is only in paractice an issue fo the tests that
        # re-initializes the application for each test, we thus
        # need to remove all AdapterResourceHandlers from the
        # factory.
        ResourceHandlerFactory.unregister_all()

        logging.getLogger("MX3.HWR").info("Starting MXCuBE-Web...")
        MXCUBEApplication.server = server
        MXCUBEApplication.ALLOW_REMOTE = allow_remote
        MXCUBEApplication.TIMEOUT_GIVES_CONTROL = ra_timeout
        MXCUBEApplication.CONFIG = cfg
        MXCUBEApplication.mxcubecore = HardwareObjectAdapterManager(MXCUBEApplication)

        MXCUBEApplication.mxcubecore.init()

        MXCUBEApplication.init_sample_video(
            _format=cfg.app.VIDEO_FORMAT,
            port=cfg.app.VIDEO_STREAM_PORT,
        )

        MXCUBEApplication.init_logging(log_fpath, log_level, enabled_logger_list)

        _UserManagerCls = import_component(
            cfg.app.usermanager, package="components.user"
        )

        MXCUBEApplication.queue = Queue(MXCUBEApplication, {})
        MXCUBEApplication.lims = Lims(MXCUBEApplication, {})
        MXCUBEApplication.usermanager = _UserManagerCls(
            MXCUBEApplication, cfg.app.usermanager
        )
        MXCUBEApplication.chat = Chat(MXCUBEApplication, {})
        MXCUBEApplication.sample_changer = SampleChanger(MXCUBEApplication, {})
        MXCUBEApplication.sample_view = SampleView(MXCUBEApplication, {})
        MXCUBEApplication.workflow = Workflow(MXCUBEApplication, {})
        MXCUBEApplication.harvester = Harvester(MXCUBEApplication, {})
        MXCUBEApplication.log = Log(MXCUBEApplication, {})

        MXCUBEApplication.init_signal_handlers()
        # Install server-side UI state storage
        MXCUBEApplication.init_state_storage()

        msg = "MXCuBE initialized, it took %.1f seconds" % (
            time.time() - MXCUBEApplication.t0
        )
        logging.getLogger("MX3.HWR").info(msg)

    @staticmethod
    def init_sample_video(_format, port):
        """
        Initializes video streaming
        :return: None
        """
        try:
            HWR.beamline.sample_view.camera.start_streaming(_format=_format, port=port)
        except Exception as ex:
            msg = "Could not initialize video, error was: "
            msg += str(ex)
            logging.getLogger("HWR").info(msg)

    @staticmethod
    def init_signal_handlers():
        """
        Connects the signal handlers defined in routes/signals.py to the
        corresponding signals/events
        """
        MXCUBEApplication.queue.init_signals(HWR.beamline.queue_model)
        MXCUBEApplication.sample_view.init_signals()
        MXCUBEApplication.sample_changer.init_signals()
        MXCUBEApplication.harvester.init_signals()

    @staticmethod
    def _get_graylog_handler(config, log_level):
        if graypy is None:
            return None
        server_cfg = getattr(config, "server", None)
        graylog_host = getattr(server_cfg, "GRAYLOG_HOST", None)
        graylog_port = getattr(server_cfg, "GRAYLOG_PORT", None)
        if graylog_host and graylog_port:
            try:
                handler = graypy.GELFUDPHandler(graylog_host, graylog_port)
            except Exception as ex:
                msg = "Graylog handler could not be initialized: " + str(ex)
                logging.getLogger("HWR").info(msg)
            else:
                handler.setLevel(log_level)
                return handler
        return None

    @staticmethod
    def init_logging(log_file, log_level, enabled_logger_list):
        """
        :param str log_file: Path to log file

        :return: None
        """
        removeLoggingHandlers()

        fmt = "%(asctime)s |%(name)-7s|%(levelname)-7s| %(message)s"
        console_formatter = ColorFormatter(fmt)
        file_formatter = logging.Formatter(fmt)

        if log_file:
            if not os.path.isfile(log_file):
                fpt = open(log_file, "w")
                fpt.write(" ")
                fpt.close()
            Path(log_file).touch()

            log_file_handler = TimedRotatingFileHandler(
                log_file, when="midnight", backupCount=7
            )
            log_file_handler.setFormatter(file_formatter)

            uilog_file = f"{log_file[:-4]}_ui.log"
            if not os.path.isfile(uilog_file):
                fpt = open(uilog_file, "w")
                fpt.write(" ")
                fpt.close()
            Path(uilog_file).touch()

            uilog_file_handler = TimedRotatingFileHandler(
                uilog_file, when="midnight", backupCount=7
            )
            uilog_file_handler.setFormatter(file_formatter)

        log_level = "INFO" if not log_level else log_level.upper()

        custom_log_handler = MX3LoggingHandler(MXCUBEApplication.server)
        custom_log_handler.setLevel(log_level)
        custom_log_handler.setFormatter(file_formatter)
        gelf_handler = MXCUBEApplication._get_graylog_handler(
            MXCUBEApplication.CONFIG, log_level
        )

        _loggers = {
            "hwr_logger": logging.getLogger("HWR"),
            "mx3_hwr_logger": logging.getLogger("MX3.HWR"),
            "user_logger": logging.getLogger("user_level_log"),
            "queue_logger": logging.getLogger("queue_exec"),
            "mx3_ui_logger": logging.getLogger("MX3.UI"),
            "csp_logger": logging.getLogger("csp"),
        }

        stdout_log_handler = StreamHandler(sys.stdout)
        stdout_log_handler.setFormatter(console_formatter)

        for logger_name, logger in _loggers.items():
            if logger_name in enabled_logger_list:
                logger.addHandler(custom_log_handler)
                logger.addHandler(stdout_log_handler)
                if gelf_handler:
                    logger.addHandler(gelf_handler)
                logger.setLevel(log_level)

                if log_file and "mx3_ui" in logger_name:
                    logger.addHandler(uilog_file_handler)
                elif log_file:
                    logger.addHandler(log_file_handler)

                logger.propagate = False
            else:
                logger.disabled = True

    @staticmethod
    def init_state_storage():
        """
        Set up of server side state storage, the UI state of the client is
        stored on the server
        """
        from mxcubeweb import state_storage

        state_storage.init()

    @staticmethod
    def get_ui_properties():
        # Add type information to each component retrieved from the beamline adapter
        # (either via config or via mxcubecore.beamline)

        for _id, section in MXCUBEApplication.CONFIG.app.ui_properties:
            if section and hasattr(section, "components"):
                for component in section.components:
                    # Check that the component, if it's a UIComponentModel, corresponds
                    # to a HardwareObjects that is available and that it can be
                    # adapted.
                    if isinstance(component, UIComponentModel):
                        try:
                            mxcore = MXCUBEApplication.mxcubecore
                            adapter = mxcore.get_adapter(component.attribute)
                            adapter_cls_name = type(adapter).__name__
                            value_type = adapter.adapter_type
                        except AttributeError:
                            msg = (
                                f"{component.attribute} not accessible via Beamline"
                                " object. "
                            )
                            msg += (
                                f"Verify that beamline.{component.attribute} is valid"
                                " and/or "
                            )
                            msg += f"{component.attribute} accessible via get_role "
                            msg += "check ui.yaml configuration file. "
                            msg += "(attribute will NOT be available in UI)"
                            logging.getLogger("HWR").warning(msg)
                            adapter_cls_name = ""
                            value_type = ""
                        else:
                            adapter_cls_name = adapter_cls_name.replace("Adapter", "")

                        if not component.object_type:
                            component.object_type = adapter_cls_name

                        if not component.value_type:
                            component.value_type = value_type

        return {
            key: value.dict()
            for (
                key,
                value,
            ) in MXCUBEApplication.CONFIG.app.ui_properties
            if value
        }
